"""
Models following scikit-learn's estimator API.
"""
from sklearn.base import BaseEstimator

from . import algorithms, families
from .utils import (
    accuracy_score,
    add_intercept,
    dot,
    exp,
    is_dask_array_sparse,
    mean_squared_error,
    poisson_deviance,
    sigmoid,
)


class _GLM(BaseEstimator):
    """Base estimator for Generalized Linear Models

    You should not use this class directly, you should use one of its subclasses
    instead.

    This class should be subclassed and paired with a GLM Family object like
    Logistic, Linear, Poisson, etc. to form an estimator.

    See Also
    --------
    LinearRegression
    LogisticRegression
    PoissonRegression
    """

    @property
    def family(self):
        """The family for which this is the estimator"""

    def __init__(
        self,
        fit_intercept=True,
        solver="admm",
        regularizer="l2",
        max_iter=100,
        tol=1e-4,
        lamduh=1.0,
        rho=1,
        over_relax=1,
        abstol=1e-4,
        reltol=1e-2,
    ):
        self.fit_intercept = fit_intercept
        self.solver = solver
        self.regularizer = regularizer
        self.max_iter = max_iter
        self.tol = tol
        self.lamduh = lamduh
        self.rho = rho
        self.over_relax = over_relax
        self.abstol = abstol
        self.reltol = reltol

        self.coef_ = None
        self.intercept_ = None
        self._coef = None  # coef, maybe with intercept

        fit_kwargs = {"max_iter", "tol", "family"}

        if solver == "admm":
            fit_kwargs.discard("tol")
            fit_kwargs.update(
                {"regularizer", "lamduh", "rho", "over_relax", "abstol", "reltol"}
            )
        elif solver == "proximal_grad" or solver == "lbfgs":
            fit_kwargs.update({"regularizer", "lamduh"})

        self._fit_kwargs = {k: getattr(self, k) for k in fit_kwargs}

    def fit(self, X, y=None):
        X_ = self._maybe_add_intercept(X)
        fit_kwargs = dict(self._fit_kwargs)
        if is_dask_array_sparse(X):
            fit_kwargs["normalize"] = False

        self._coef = algorithms._solvers[self.solver](X_, y, **fit_kwargs)

        if self.fit_intercept:
            self.coef_ = self._coef[:-1]
            self.intercept_ = self._coef[-1]
        else:
            self.coef_ = self._coef
        return self

    def _maybe_add_intercept(self, X):
        if self.fit_intercept:
            return add_intercept(X)
        else:
            return X


class LogisticRegression(_GLM):
    """
    Estimator for logistic regression.

    Parameters
    ----------
    fit_intercept : bool, default True
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the decision function.
    solver : {'admm', 'gradient_descent', 'newton', 'lbfgs', 'proximal_grad'}
        Solver to use. See :ref:`api.algorithms` for details
    regularizer : {'l1', 'l2'}
        Regularizer to use. See :ref:`api.regularizers` for details.
        Only used with ``admm``, ``lbfgs``, and ``proximal_grad`` solvers.
    max_iter : int, default 100
        Maximum number of iterations taken for the solvers to converge
    tol : float, default 1e-4
        Tolerance for stopping criteria. Ignored for ``admm`` solver
    lambduh : float, default 1.0
        Only used with ``admm``, ``lbfgs`` and ``proximal_grad`` solvers.
    rho, over_relax, abstol, reltol : float
        Only used with the ``admm`` solver.

    Attributes
    ----------
    coef_ : array, shape (n_classes, n_features)
        The learned value for the model's coefficients
    intercept_ : float of None
        The learned value for the intercept, if one was added
        to the model

    Examples
    --------
    >>> from dask_glm.datasets import make_classification
    >>> X, y = make_classification()
    >>> est = LogisticRegression()
    >>> est.fit(X, y)
    >>> est.predict(X)
    >>> est.predict_proba(X)
    >>> est.score(X, y)
    """

    family = families.Logistic

    def predict(self, X):
        return self.predict_proba(X) > 0.5  # TODO: verify, multiclass broken

    def predict_proba(self, X):
        X_ = self._maybe_add_intercept(X)
        return sigmoid(dot(X_, self._coef))

    def score(self, X, y):
        return accuracy_score(y, self.predict(X))


class LinearRegression(_GLM):
    """
    Estimator for a linear model using Ordinary Least Squares.

    Parameters
    ----------
    fit_intercept : bool, default True
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the decision function.
    solver : {'admm', 'gradient_descent', 'newton', 'lbfgs', 'proximal_grad'}
        Solver to use. See :ref:`api.algorithms` for details
    regularizer : {'l1', 'l2'}
        Regularizer to use. See :ref:`api.regularizers` for details.
        Only used with ``admm`` and ``proximal_grad`` solvers.
    max_iter : int, default 100
        Maximum number of iterations taken for the solvers to converge
    tol : float, default 1e-4
        Tolerance for stopping criteria. Ignored for ``admm`` solver
    lambduh : float, default 1.0
        Only used with ``admm`` and ``proximal_grad`` solvers
    rho, over_relax, abstol, reltol : float
        Only used with the ``admm`` solver.

    Attributes
    ----------
    coef_ : array, shape (n_classes, n_features)
        The learned value for the model's coefficients
    intercept_ : float of None
        The learned value for the intercept, if one was added
        to the model

    Examples
    --------
    >>> from dask_glm.datasets import make_regression
    >>> X, y = make_regression()
    >>> est = LinearRegression()
    >>> est.fit(X, y)
    >>> est.predict(X)
    >>> est.score(X, y)
    """

    family = families.Normal

    def predict(self, X):
        X_ = self._maybe_add_intercept(X)
        return dot(X_, self._coef)

    def score(self, X, y):
        return mean_squared_error(y, self.predict(X))


class PoissonRegression(_GLM):
    """
    Estimator for Poisson Regression.

    Parameters
    ----------
    fit_intercept : bool, default True
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the decision function.
    solver : {'admm', 'gradient_descent', 'newton', 'lbfgs', 'proximal_grad'}
        Solver to use. See :ref:`api.algorithms` for details
    regularizer : {'l1', 'l2'}
        Regularizer to use. See :ref:`api.regularizers` for details.
        Only used with ``admm``, ``lbfgs``, and ``proximal_grad`` solvers.
    max_iter : int, default 100
        Maximum number of iterations taken for the solvers to converge
    tol : float, default 1e-4
        Tolerance for stopping criteria. Ignored for ``admm`` solver
    lambduh : float, default 1.0
        Only used with ``admm``, ``lbfgs`` and ``proximal_grad`` solvers.
    rho, over_relax, abstol, reltol : float
        Only used with the ``admm`` solver.

    Attributes
    ----------
    coef_ : array, shape (n_classes, n_features)
        The learned value for the model's coefficients
    intercept_ : float of None
        The learned value for the intercept, if one was added
        to the model

    Examples
    --------
    >>> from dask_glm.datasets import make_poisson
    >>> X, y = make_poisson()
    >>> est = PoissonRegression()
    >>> est.fit(X, y)
    >>> est.predict(X)
    >>> est.get_deviance(X, y)
    """

    family = families.Poisson

    def predict(self, X):
        X_ = self._maybe_add_intercept(X)
        return exp(dot(X_, self._coef))

    def get_deviance(self, X, y):
        return poisson_deviance(y, self.predict(X))



import numpy as np
import numpy.testing as npt
import pytest

from dask_glm import regularizers as regs


@pytest.mark.parametrize(
    "func,args",
    [("f", [0]), ("gradient", [0]), ("hessian", [0]), ("proximal_operator", [0, 1])],
)
def test_base_class_raises_notimplementederror(func, args):
    with pytest.raises(NotImplementedError):
        getattr(regs.Regularizer(), func)(*args)


class FooRegularizer(regs.Regularizer):
    def f(self, beta):
        return beta + 1

    def gradient(self, beta):
        return beta + 1

    def hessian(self, beta):
        return beta + 1


@pytest.mark.parametrize("func", ["add_reg_f", "add_reg_grad", "add_reg_hessian"])
def test_add_reg_funcs(func):
    def foo(x):
        return x**2

    new_func = getattr(FooRegularizer(), func)(foo, 1)
    assert callable(new_func)
    assert new_func(2) == 7


def test_regularizer_get_passes_through_instance():
    x = FooRegularizer()
    assert regs.Regularizer.get(x) == x


def test_regularizer_get_unnamed_raises():
    with pytest.raises(KeyError):
        regs.Regularizer.get("foo")


def test_regularizer_gets_from_name():
    class Foo(regs.Regularizer):
        name = "foo"

    assert isinstance(regs.Regularizer.get("foo"), Foo)


@pytest.mark.parametrize(
    "beta,expected", [(np.array([0, 0, 0]), 0), (np.array([1, 2, 3]), 7)]
)
def test_l2_function(beta, expected):
    assert regs.L2().f(beta) == expected


@pytest.mark.parametrize("beta", [np.array([0, 0, 0]), np.array([1, 2, 3])])
def test_l2_gradient(beta):
    npt.assert_array_equal(regs.L2().gradient(beta), beta)


@pytest.mark.parametrize("beta", [np.array([0, 0, 0]), np.array([1, 2, 3])])
def test_l2_hessian(beta):
    npt.assert_array_equal(regs.L2().hessian(beta), np.eye(len(beta)))


@pytest.mark.parametrize(
    "beta,expected",
    [
        (np.array([0, 0, 0]), np.array([0, 0, 0])),
        (np.array([1, 2, 3]), np.array([0.5, 1, 1.5])),
    ],
)
def test_l2_proximal_operator(beta, expected):
    npt.assert_array_equal(regs.L2().proximal_operator(beta, 1), expected)


@pytest.mark.parametrize(
    "beta,expected", [(np.array([0, 0, 0]), 0), (np.array([-1, 2, 3]), 6)]
)
def test_l1_function(beta, expected):
    assert regs.L1().f(beta) == expected


@pytest.mark.parametrize(
    "beta,expected",
    [
        (np.array([1, 2, 3]), np.array([1, 1, 1])),
        (np.array([-1, 2, 3]), np.array([-1, 1, 1])),
    ],
)
def test_l1_gradient(beta, expected):
    npt.assert_array_equal(regs.L1().gradient(beta), expected)


@pytest.mark.parametrize(
    "beta",
    [np.array([0.00000001, 1, 2]), np.array([-0.00000001, 1, 2]), np.array([0, 0, 0])],
)
def test_l1_gradient_raises_near_zero(beta):
    with pytest.raises(ValueError):
        regs.L1().gradient(beta)


def test_l1_hessian():
    npt.assert_array_equal(
        regs.L1().hessian(np.array([1, 2])), np.array([[0, 0], [0, 0]])
    )


def test_l1_hessian_raises():
    with pytest.raises(ValueError):
        regs.L1().hessian(np.array([0, 0, 0]))


@pytest.mark.parametrize(
    "beta,expected",
    [
        (np.array([0, 0, 0]), np.array([0, 0, 0])),
        (np.array([1, 2, 3]), np.array([0, 1, 2])),
    ],
)
def test_l1_proximal_operator(beta, expected):
    npt.assert_array_equal(regs.L1().proximal_operator(beta, 1), expected)


@pytest.mark.parametrize(
    "beta,expected", [(np.array([0, 0, 0]), 0), (np.array([1, 2, 3]), 6.5)]
)
def test_elastic_net_function(beta, expected):
    assert regs.ElasticNet().f(beta) == expected


def test_elastic_net_function_zero_weight_is_l2():
    beta = np.array([1, 2, 3])
    assert regs.ElasticNet(weight=0).f(beta) == regs.L2().f(beta)


def test_elastic_net_function_zero_weight_is_l1():
    beta = np.array([1, 2, 3])
    assert regs.ElasticNet(weight=1).f(beta) == regs.L1().f(beta)


def test_elastic_net_gradient():
    beta = np.array([1, 2, 3])
    npt.assert_array_equal(
        regs.ElasticNet(weight=0.5).gradient(beta), np.array([1, 1.5, 2])
    )


def test_elastic_net_gradient_zero_weight_is_l2():
    beta = np.array([1, 2, 3])
    npt.assert_array_equal(
        regs.ElasticNet(weight=0).gradient(beta), regs.L2().gradient(beta)
    )


def test_elastic_net_gradient_zero_weight_is_l1():
    beta = np.array([1, 2, 3])
    npt.assert_array_equal(
        regs.ElasticNet(weight=1).gradient(beta), regs.L1().gradient(beta)
    )


def test_elastic_net_hessian():
    beta = np.array([1, 2, 3])
    npt.assert_array_equal(
        regs.ElasticNet(weight=0.5).hessian(beta),
        np.eye(len(beta)) * regs.ElasticNet().weight,
    )


def test_elastic_net_hessian_raises():
    with pytest.raises(ValueError):
        regs.ElasticNet(weight=0.5).hessian(np.array([0, 1, 2]))


def test_elastic_net_proximal_operator():
    beta = np.array([1, 2, 3])
    npt.assert_array_equal(regs.ElasticNet(weight=0.5).proximal_operator(beta, 1), beta)



import dask
import dask.array as da
import dask.multiprocessing
import numpy as np
import pytest
from dask import persist

from dask_glm.algorithms import admm, gradient_descent, lbfgs, newton, proximal_grad
from dask_glm.families import Logistic, Normal, Poisson
from dask_glm.regularizers import Regularizer
from dask_glm.utils import make_y, maybe_to_cupy, sigmoid, to_dask_cupy_array_xy


def add_l1(f, lam):
    def wrapped(beta, X, y):
        return f(beta, X, y) + lam * (np.abs(beta)).sum()

    return wrapped


def make_intercept_data(N, p, seed=20009):
    """Given the desired number of observations (N) and
    the desired number of variables (p), creates
    random logistic data to test on."""

    # set the seeds
    da.random.seed(seed)
    np.random.seed(seed)

    X = np.random.random((N, p + 1))
    col_sums = X.sum(axis=0)
    X = X / col_sums[None, :]
    X[:, p] = 1
    X = da.from_array(X, chunks=(N / 5, p + 1))
    y = make_y(X, beta=np.random.random(p + 1))

    return X, y


@pytest.mark.parametrize("opt", [lbfgs, newton, gradient_descent])
@pytest.mark.parametrize(
    "N, p, seed,", [(100, 2, 20009), (250, 12, 90210), (95, 6, 70605)]
)
@pytest.mark.parametrize("is_cupy", [True, False])
def test_methods(N, p, seed, opt, is_cupy):
    X, y = make_intercept_data(N, p, seed=seed)

    if is_cupy:
        cupy = pytest.importorskip("cupy")
        X, y = to_dask_cupy_array_xy(X, y, cupy)

    coefs = opt(X, y)
    p = sigmoid(X.dot(coefs).compute())

    y_sum = y.compute().sum()
    p_sum = p.sum()
    assert np.isclose(y_sum, p_sum, atol=1e-1)


@pytest.mark.parametrize(
    "func,kwargs",
    [
        (newton, {"tol": 1e-5}),
        (lbfgs, {"tol": 1e-8}),
        (gradient_descent, {"tol": 1e-7}),
    ],
)
@pytest.mark.parametrize("N", [1000])
@pytest.mark.parametrize("nchunks", [1, 10])
@pytest.mark.parametrize("family", [Logistic, Normal, Poisson])
@pytest.mark.parametrize("is_cupy", [True, False])
def test_basic_unreg_descent(func, kwargs, N, nchunks, family, is_cupy):
    beta = np.random.normal(size=2)
    M = len(beta)
    X = da.random.random((N, M), chunks=(N // nchunks, M))
    y = make_y(X, beta=np.array(beta), chunks=(N // nchunks,))

    if is_cupy:
        cupy = pytest.importorskip("cupy")
        X, y = to_dask_cupy_array_xy(X, y, cupy)

    X, y = persist(X, y)

    result = func(X, y, family=family, **kwargs)
    test_vec = np.random.normal(size=2)
    test_vec = maybe_to_cupy(test_vec, X)

    opt = family.pointwise_loss(result, X, y).compute()
    test_val = family.pointwise_loss(test_vec, X, y).compute()

    assert opt < test_val


@pytest.mark.parametrize(
    "func,kwargs",
    [
        (admm, {"abstol": 1e-4}),
        (proximal_grad, {"tol": 1e-7}),
    ],
)
@pytest.mark.parametrize("N", [1000])
@pytest.mark.parametrize("nchunks", [1, 10])
@pytest.mark.parametrize("family", [Logistic, Normal, Poisson])
@pytest.mark.parametrize("lam", [0.01, 1.2, 4.05])
@pytest.mark.parametrize("reg", [r() for r in Regularizer.__subclasses__()])
@pytest.mark.parametrize("is_cupy", [True, False])
def test_basic_reg_descent(func, kwargs, N, nchunks, family, lam, reg, is_cupy):
    beta = np.random.normal(size=2)
    M = len(beta)
    X = da.random.random((N, M), chunks=(N // nchunks, M))
    y = make_y(X, beta=np.array(beta), chunks=(N // nchunks,))

    if is_cupy:
        cupy = pytest.importorskip("cupy")
        X, y = to_dask_cupy_array_xy(X, y, cupy)

    X, y = persist(X, y)

    result = func(X, y, family=family, lamduh=lam, regularizer=reg, **kwargs)
    test_vec = np.random.normal(size=2)
    test_vec = maybe_to_cupy(test_vec, X)

    f = reg.add_reg_f(family.pointwise_loss, lam)

    opt = f(result, X, y).compute()
    test_val = f(test_vec, X, y).compute()

    assert opt < test_val


@pytest.mark.parametrize(
    "func,kwargs",
    [
        (admm, {"max_iter": 2}),
        (proximal_grad, {"max_iter": 2}),
        (newton, {"max_iter": 2}),
        (gradient_descent, {"max_iter": 2}),
    ],
)
@pytest.mark.parametrize("scheduler", ["synchronous", "threading", "multiprocessing"])
@pytest.mark.parametrize("is_cupy", [True, False])
def test_determinism(func, kwargs, scheduler, is_cupy):
    X, y = make_intercept_data(1000, 10)
    if is_cupy:
        cupy = pytest.importorskip("cupy")
        X, y = to_dask_cupy_array_xy(X, y, cupy)

    with dask.config.set(scheduler=scheduler):
        a = func(X, y, **kwargs)
        b = func(X, y, **kwargs)

    assert (a == b).all()


try:
    from distributed import Client
    from distributed.utils_test import cleanup, cluster, loop_in_thread  # noqa: F401
except ImportError:
    pass
else:

    @pytest.mark.parametrize(
        "func,kwargs",
        [
            (admm, {"max_iter": 2}),
            (proximal_grad, {"max_iter": 2}),
            (newton, {"max_iter": 2}),
            (gradient_descent, {"max_iter": 2}),
        ],
    )
    def test_determinism_distributed(func, kwargs, loop_in_thread):  # noqa: F811
        with cluster() as (s, [a, b]):
            with Client(s["address"], loop=loop_in_thread) as _:
                X, y = make_intercept_data(1000, 10)

                a = func(X, y, **kwargs)
                b = func(X, y, **kwargs)

                assert (a == b).all()

    def test_broadcast_lbfgs_weight(loop_in_thread):  # noqa: F811
        with cluster() as (s, [a, b]):
            with Client(s["address"], loop=loop_in_thread) as c:
                X, y = make_intercept_data(1000, 10)
                coefs = lbfgs(X, y, dask_distributed_client=c)
                p = sigmoid(X.dot(coefs).compute())

                y_sum = y.compute().sum()
                p_sum = p.sum()
                assert np.isclose(y_sum, p_sum, atol=1e-1)



"""Optimization algorithms for solving minimizaiton problems.
"""

from __future__ import absolute_import, division, print_function

import functools

import dask
import dask.array as da
import numpy as np
from dask import compute, delayed, persist
from dask.array.utils import normalize_to_array
from scipy.optimize import fmin_l_bfgs_b

from dask_glm.families import Logistic
from dask_glm.regularizers import Regularizer
from dask_glm.utils import (
    dot,
    get_distributed_client,
    maybe_to_cupy,
    normalize,
    scatter_array,
)


def compute_stepsize_dask(
    beta,
    step,
    Xbeta,
    Xstep,
    y,
    curr_val,
    family=Logistic,
    stepSize=1.0,
    armijoMult=0.1,
    backtrackMult=0.1,
):
    """Compute the optimal stepsize

    Parameters
    ----------
    beta : array-like
    step : float
    XBeta : array-like
    Xstep : float
    y : array-like
    curr_val : float
    famlily : Family, optional
    stepSize : float, optional
    armijoMult : float, optional
    backtrackMult : float, optional

    Returns
    -------
    stepSize : float
    beta : array-like
    xBeta : array-like
    func : callable
    """

    loglike = family.loglike
    beta, step, Xbeta, Xstep, y, curr_val = persist(
        beta, step, Xbeta, Xstep, y, curr_val
    )
    obeta, oXbeta = beta, Xbeta
    (step,) = compute(step)
    steplen = (step**2).sum()
    lf = curr_val
    func = 0
    for ii in range(100):
        beta = obeta - stepSize * step
        if ii and (beta == obeta).all():
            stepSize = 0
            break

        Xbeta = oXbeta - stepSize * Xstep
        func = loglike(Xbeta, y)
        Xbeta, func = persist(Xbeta, func)

        df = lf - compute(func)[0]
        if df >= armijoMult * stepSize * steplen:
            break
        stepSize *= backtrackMult

    return stepSize, beta, Xbeta, func


@normalize
def gradient_descent(X, y, max_iter=100, tol=1e-14, family=Logistic, **kwargs):
    """
    Michael Grant's implementation of Gradient Descent.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
    y : array-like, shape (n_samples,)
    max_iter : int
        maximum number of iterations to attempt before declaring
        failure to converge
    tol : float
        Maximum allowed change from prior iteration required to
        declare convergence
    family : Family

    Returns
    -------
    beta : array-like, shape (n_features,)
    """
    loglike, gradient = family.loglike, family.gradient
    n, p = X.shape
    firstBacktrackMult = 0.1
    nextBacktrackMult = 0.5
    armijoMult = 0.1
    stepGrowth = 1.25
    stepSize = 1.0
    recalcRate = 10
    backtrackMult = firstBacktrackMult
    beta = np.zeros_like(X._meta, shape=p)

    for k in range(max_iter):
        # how necessary is this recalculation?
        if k % recalcRate == 0:
            Xbeta = X.dot(beta)
            func = loglike(Xbeta, y)

        grad = gradient(Xbeta, X, y)
        Xgradient = X.dot(grad)

        # backtracking line search
        lf = func
        stepSize, _, _, func = compute_stepsize_dask(
            beta,
            grad,
            Xbeta,
            Xgradient,
            y,
            func,
            family=family,
            backtrackMult=backtrackMult,
            armijoMult=armijoMult,
            stepSize=stepSize,
        )

        beta, stepSize, Xbeta, lf, func, grad, Xgradient = persist(
            beta, stepSize, Xbeta, lf, func, grad, Xgradient
        )

        stepSize, lf, func, grad = compute(stepSize, lf, func, grad)

        beta = (
            beta - stepSize * grad
        )  # tiny bit of repeat work here to avoid communication
        Xbeta = Xbeta - stepSize * Xgradient

        if stepSize == 0:
            break

        df = lf - func
        df /= max(func, lf)

        if df < tol:
            break
        stepSize *= stepGrowth
        backtrackMult = nextBacktrackMult

    return beta


@normalize
def newton(X, y, max_iter=50, tol=1e-8, family=Logistic, **kwargs):
    """Newton's Method for Logistic Regression.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
    y : array-like, shape (n_samples,)
    max_iter : int
        maximum number of iterations to attempt before declaring
        failure to converge
    tol : float
        Maximum allowed change from prior iteration required to
        declare convergence
    family : Family

    Returns
    -------
    beta : array-like, shape (n_features,)
    """
    gradient, hessian = family.gradient, family.hessian
    n, p = X.shape
    beta = np.zeros_like(X._meta, shape=p)
    Xbeta = dot(X, beta)

    iter_count = 0
    converged = False

    while not converged:
        beta_old = beta

        # should this use map_blocks()?
        hess = hessian(Xbeta, X)
        grad = gradient(Xbeta, X, y)

        hess, grad = da.compute(hess, grad)

        # should this be dask or numpy?
        # currently uses Python 3 specific syntax
        step, _, _, _ = np.linalg.lstsq(hess, grad)
        beta = beta_old - step

        iter_count += 1

        # should change this criterion
        coef_change = np.absolute(beta_old - beta)
        converged = (not np.any(coef_change > tol)) or (iter_count > max_iter)

        if not converged:
            Xbeta = dot(X, beta)  # numpy -> dask converstion of beta

    return beta


@normalize
def admm(
    X,
    y,
    regularizer="l1",
    lamduh=0.1,
    rho=1,
    over_relax=1,
    max_iter=250,
    abstol=1e-4,
    reltol=1e-2,
    family=Logistic,
    **kwargs
):
    """
    Alternating Direction Method of Multipliers

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
    y : array-like, shape (n_samples,)
    regularizer : str or Regularizer
    lamduh : float
    rho : float
    over_relax : FLOAT
    max_iter : int
        maximum number of iterations to attempt before declaring
        failure to converge
    abstol, reltol : float
    family : Family

    Returns
    -------
    beta : array-like, shape (n_features,)
    """
    pointwise_loss = family.pointwise_loss
    pointwise_gradient = family.pointwise_gradient
    regularizer = Regularizer.get(regularizer)

    def create_local_gradient(func):
        @functools.wraps(func)
        def wrapped(beta, X, y, z, u, rho):
            beta = maybe_to_cupy(beta, X)
            z = maybe_to_cupy(z, X)
            u = maybe_to_cupy(u, X)
            res = func(beta, X, y) + rho * (beta - z + u)
            return normalize_to_array(res)

        return wrapped

    def create_local_f(func):
        @functools.wraps(func)
        def wrapped(beta, X, y, z, u, rho):
            beta = maybe_to_cupy(beta, X)
            z = maybe_to_cupy(z, X)
            u = maybe_to_cupy(u, X)
            res = func(beta, X, y) + (rho / 2) * np.dot(beta - z + u, beta - z + u)
            return normalize_to_array(res)

        return wrapped

    f = create_local_f(pointwise_loss)
    fprime = create_local_gradient(pointwise_gradient)

    nchunks = getattr(X, "npartitions", 1)
    # nchunks = X.npartitions
    (n, p) = X.shape
    # XD = X.to_delayed().flatten().tolist()
    # yD = y.to_delayed().flatten().tolist()
    if isinstance(X, da.Array):
        XD = X.rechunk((None, X.shape[-1])).to_delayed().flatten().tolist()
    else:
        XD = [X]
    if isinstance(y, da.Array):
        yD = y.rechunk((None, y.shape[-1])).to_delayed().flatten().tolist()
    else:
        yD = [y]

    z = np.zeros(p)
    u = np.array([np.zeros(p) for i in range(nchunks)])
    betas = np.array([np.ones(p) for i in range(nchunks)])

    for k in range(max_iter):
        # x-update step
        new_betas = [
            delayed(local_update)(xx, yy, bb, z, uu, rho, f=f, fprime=fprime)
            for xx, yy, bb, uu in zip(XD, yD, betas, u)
        ]
        new_betas = np.array(da.compute(*new_betas))

        beta_hat = over_relax * new_betas + (1 - over_relax) * z

        #  z-update step
        zold = z.copy()
        ztilde = np.mean(beta_hat + np.array(u), axis=0)
        z = regularizer.proximal_operator(ztilde, lamduh / (rho * nchunks))

        # u-update step
        u += beta_hat - z

        # check for convergence
        primal_res = np.linalg.norm(new_betas - z)
        dual_res = np.linalg.norm(rho * (z - zold))

        eps_pri = np.sqrt(p * nchunks) * abstol + reltol * np.maximum(
            np.linalg.norm(new_betas), np.sqrt(nchunks) * np.linalg.norm(z)
        )
        eps_dual = np.sqrt(p * nchunks) * abstol + reltol * np.linalg.norm(rho * u)

        if primal_res < eps_pri and dual_res < eps_dual:
            break

    return maybe_to_cupy(z, X)


def local_update(X, y, beta, z, u, rho, f, fprime, solver=fmin_l_bfgs_b):
    beta = beta.ravel()
    u = u.ravel()
    z = z.ravel()
    solver_args = (X, y, z, u, rho)
    beta, f, d = solver(
        f, beta, fprime=fprime, args=solver_args, maxiter=200, maxfun=250
    )

    return beta


@normalize
def lbfgs(
    X,
    y,
    regularizer=None,
    lamduh=1.0,
    max_iter=100,
    tol=1e-4,
    family=Logistic,
    verbose=False,
    **kwargs
):
    """L-BFGS solver using scipy.optimize implementation

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
    y : array-like, shape (n_samples,)
    regularizer : str or Regularizer
    lamduh : float
    max_iter : int
        maximum number of iterations to attempt before declaring
        failure to converge
    tol : float
        Maximum allowed change from prior iteration required to
        declare convergence
    family : Family
    verbose : bool, default False
        whether to print diagnostic information during convergence

    Returns
    -------
    beta : array-like, shape (n_features,)
    """
    dask_distributed_client = get_distributed_client()
    pointwise_loss = family.pointwise_loss
    pointwise_gradient = family.pointwise_gradient
    if regularizer is not None:
        regularizer = Regularizer.get(regularizer)
        pointwise_loss = regularizer.add_reg_f(pointwise_loss, lamduh)
        pointwise_gradient = regularizer.add_reg_grad(pointwise_gradient, lamduh)

    n, p = X.shape
    beta0 = np.zeros(p)

    def compute_loss_grad(beta, X, y):
        beta = maybe_to_cupy(beta, X)
        scatter_beta = (
            scatter_array(beta, dask_distributed_client)
            if dask_distributed_client
            else beta
        )
        loss_fn = pointwise_loss(scatter_beta, X, y)
        gradient_fn = pointwise_gradient(scatter_beta, X, y)
        loss, gradient = compute(loss_fn, gradient_fn)
        return normalize_to_array(loss), normalize_to_array(gradient.copy())

    with dask.config.set(fuse_ave_width=0):  # optimizations slows this down
        beta, loss, info = fmin_l_bfgs_b(
            compute_loss_grad,
            beta0,
            fprime=None,
            args=(X, y),
            iprint=(verbose > 0) - 1,
            pgtol=tol,
            maxiter=max_iter,
        )
    beta = maybe_to_cupy(beta, X)
    return beta


@normalize
def proximal_grad(
    X,
    y,
    regularizer="l1",
    lamduh=0.1,
    family=Logistic,
    max_iter=100,
    tol=1e-8,
    **kwargs
):
    """
    Proximal Gradient Method

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
    y : array-like, shape (n_samples,)
    regularizer : str or Regularizer
    lamduh : float
    max_iter : int
        maximum number of iterations to attempt before declaring
        failure to converge
    tol : float
        Maximum allowed change from prior iteration required to
        declare convergence
    family : Family

    Returns
    -------
    beta : array-like, shape (n_features,)
    """
    n, p = X.shape
    firstBacktrackMult = 0.1
    nextBacktrackMult = 0.5
    stepGrowth = 1.25
    stepSize = 1.0
    recalcRate = 10
    backtrackMult = firstBacktrackMult
    beta = np.zeros_like(X._meta, shape=p)
    regularizer = Regularizer.get(regularizer)

    for k in range(max_iter):
        # Compute the gradient
        if k % recalcRate == 0:
            Xbeta = X.dot(beta)
            func = family.loglike(Xbeta, y)

        gradient = family.gradient(Xbeta, X, y)

        Xbeta, func, gradient = persist(Xbeta, func, gradient)

        obeta = beta

        # Compute the step size
        lf = func
        for ii in range(100):
            beta = regularizer.proximal_operator(
                -stepSize * gradient + obeta, stepSize * lamduh
            )
            Xbeta = X.dot(beta)

            Xbeta, beta = persist(Xbeta, beta)

            func = family.loglike(Xbeta, y)
            func = persist(func)[0]
            func = compute(func)[0]
            df = lf - func
            if df > 0:
                break
            stepSize *= backtrackMult
        if stepSize == 0:
            break
        df /= max(func, lf)
        if df < tol:
            break
        stepSize *= stepGrowth
        backtrackMult = nextBacktrackMult

    # L2-regularization returned a dask-array
    try:
        return beta.compute()
    except AttributeError:
        return beta


_solvers = {
    "admm": admm,
    "gradient_descent": gradient_descent,
    "newton": newton,
    "lbfgs": lbfgs,
    "proximal_grad": proximal_grad,
}



from __future__ import absolute_import, division, print_function

import inspect
import sys
from functools import wraps

import dask.array as da
import numpy as np
import sparse
from dask.distributed import get_client
from multipledispatch import dispatch


def normalize(algo):
    @wraps(algo)
    def normalize_inputs(X, y, *args, **kwargs):
        normalize = kwargs.pop("normalize", True)
        if normalize:
            mean, std = da.compute(X.mean(axis=0), X.std(axis=0))
            mean, std = mean.copy(), std.copy()  # in case they are read-only
            intercept_idx = np.where(std == 0)
            if len(intercept_idx[0]) > 1:
                raise ValueError("Multiple constant columns detected!")
            mean[intercept_idx] = 0
            std[intercept_idx] = 1
            mean = (
                mean
                if len(intercept_idx[0])
                else np.zeros_like(X._meta, shape=mean.shape)
            )
            Xn = (X - mean) / std
            out = algo(Xn, y, *args, **kwargs).copy()
            i_adj = np.sum(out * mean / std)
            out[intercept_idx] -= i_adj
            return out / std
        else:
            return algo(X, y, *args, **kwargs)

    return normalize_inputs


def sigmoid(x):
    """Sigmoid function of x."""
    return 1 / (1 + exp(-x))


@dispatch(object)
def exp(A):
    return np.exp(A)


@dispatch(float)
def exp(A):
    return np.exp(A)


@dispatch(np.ndarray)
def exp(A):
    return np.exp(A)


@dispatch(da.Array)
def exp(A):
    return da.exp(A)


@dispatch(object)
def absolute(A):
    return abs(A)


@dispatch(np.ndarray)
def absolute(A):
    return np.absolute(A)


@dispatch(da.Array)
def absolute(A):
    return da.absolute(A)


@dispatch(object)
def sign(A):
    return A.sign()


@dispatch(np.ndarray)
def sign(A):
    return np.sign(A)


@dispatch(da.Array)
def sign(A):
    return da.sign(A)


@dispatch(object)
def log1p(A):
    return np.log1p(A)


@dispatch(np.ndarray)
def log1p(A):
    return np.log1p(A)


@dispatch(da.Array)
def log1p(A):
    return da.log1p(A)


@dispatch(object, object)
def dot(A, B):
    if isinstance(A, da.Array) or isinstance(B, da.Array):
        A = da.asarray(A, chunks=A.shape)
        B = da.asarray(B, chunks=B.shape)
    return np.dot(A, B)


@dispatch(object)
def sum(A):
    return A.sum()


def is_dask_array_sparse(X):
    """
    Check using _meta if a dask array contains sparse arrays
    """
    return isinstance(X._meta, sparse.SparseArray)


@dispatch(np.ndarray)
def add_intercept(X):
    return np.concatenate([X, np.ones((X.shape[0], 1))], axis=1)


@dispatch(sparse.SparseArray)
def add_intercept(X):
    return sparse.concatenate([X, sparse.COO(np.ones((X.shape[0], 1)))], axis=1)


@dispatch(da.Array)
def add_intercept(X):
    if np.isnan(np.sum(X.shape)):
        raise NotImplementedError(
            "Can not add intercept to array with " "unknown chunk shape"
        )
    j, k = X.chunks
    o = da.ones_like(X, shape=(X.shape[0], 1), chunks=(j, 1))
    if is_dask_array_sparse(X):
        o = o.map_blocks(sparse.COO)
    # TODO: Needed this `.rechunk` for the solver to work
    # Is this OK / correct?
    X_i = da.concatenate([X, o], axis=1).rechunk((j, (k[0] + 1,)))
    return X_i


def make_y(X, beta=np.array([1.5, -3]), chunks=2):
    n, p = X.shape
    z0 = X.dot(beta)
    y = da.random.random(z0.shape, chunks=z0.chunks) < sigmoid(z0)
    return y


def mean_squared_error(y_true, y_pred):
    return ((y_true - y_pred) ** 2).mean()


def accuracy_score(y_true, y_pred):
    return (y_true == y_pred).mean()


def poisson_deviance(y_true, y_pred):
    return 2 * (y_true * log1p(y_true / y_pred) - (y_true - y_pred)).sum()


try:
    import sparse
except ImportError:
    pass
else:

    @dispatch(sparse.COO)
    def exp(x):
        return np.exp(x.todense())


def package_of(obj):
    """Return package containing object's definition

    Or return None if not found
    """
    # http://stackoverflow.com/questions/43462701/get-package-of-python-object/43462865#43462865
    mod = inspect.getmodule(obj)
    if not mod:
        return
    base, _sep, _stem = mod.__name__.partition(".")
    return sys.modules[base]


def scatter_array(arr, dask_client):
    """Scatter a large numpy array into workers
    Return the equivalent dask array
    """
    future_arr = dask_client.scatter(arr)
    return da.from_delayed(
        future_arr, shape=arr.shape, dtype=arr.dtype, meta=np.zeros_like(arr, shape=())
    )


def get_distributed_client():
    try:
        return get_client()
    except ValueError:
        return None


def maybe_to_cupy(beta, X):
    """convert beta, a numpy array, to a cupy array
    if X is a cupy array or dask cupy array
    """
    if "cupy" in str(type(X)) or hasattr(X, "_meta") and "cupy" in str(type(X._meta)):
        import cupy

        return cupy.asarray(beta)
    return beta


def to_dask_cupy_array(X, cupy):
    """convert a dask numpy array to a dask cupy array"""
    return X.map_blocks(
        lambda x: cupy.asarray(x), dtype=X.dtype, meta=cupy.asarray(X._meta)
    )


def to_dask_cupy_array_xy(X, y, cupy):
    return to_dask_cupy_array(X, cupy), to_dask_cupy_array(y, cupy)



#!/usr/bin/env python

from os.path import exists

from setuptools import setup

setup(
    name="dask-glm",
    description="Generalized Linear Models with Dask",
    url="http://github.com/dask/dask-glm/",
    maintainer="Matthew Rocklin",
    maintainer_email="mrocklin@gmail.com",
    license="BSD",
    keywords="dask,glm",
    packages=["dask_glm"],
    long_description=(open("README.rst").read() if exists("README.rst") else ""),
    install_requires=list(open("requirements.txt").read().strip().split("\n")),
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    extras_require={
        "docs": [
            "jupyter",
            "nbsphinx",
            "notebook",
            "numpydoc",
            "sphinx",
            "sphinx_rtd_theme",
        ]
    },
    zip_safe=False,
)



import dask
import pytest

from dask_glm.datasets import make_classification, make_poisson, make_regression
from dask_glm.estimators import LinearRegression, LogisticRegression, PoissonRegression
from dask_glm.regularizers import Regularizer
from dask_glm.utils import to_dask_cupy_array_xy


@pytest.fixture(params=[r() for r in Regularizer.__subclasses__()])
def solver(request):
    """Parametrized fixture for all the solver names"""
    return request.param


@pytest.fixture(params=[r() for r in Regularizer.__subclasses__()])
def regularizer(request):
    """Parametrized fixture for all the regularizer names"""
    return request.param


class DoNothingTransformer(object):
    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X

    def fit_transform(self, X, y=None):
        return X

    def get_params(self, deep=True):
        return {}


X, y = make_classification()


def test_lr_init(solver):
    LogisticRegression(solver=solver)


def test_pr_init(solver):
    PoissonRegression(solver=solver)


def _maybe_skip_sparse_error(fit_intercept, is_sparse, is_cupy):
    if fit_intercept and is_sparse and not is_cupy:
        msg = (
            "ValueError: This operation requires consistent fill-values, "
            "but argument 1 had a fill value of 1.0, which is different "
            "from a fill_value of 0.0 in the first argument"
        )
        pytest.xfail(f"TODO: {msg}")


@pytest.mark.parametrize("fit_intercept", [True, False])
@pytest.mark.parametrize(
    "is_sparse,is_cupy", [(True, False), (False, False), (False, True)]
)
def test_fit(fit_intercept, is_sparse, is_cupy):
    _maybe_skip_sparse_error(fit_intercept, is_sparse, is_cupy)

    X, y = make_classification(
        n_samples=100, n_features=5, chunksize=10, is_sparse=is_sparse
    )

    if is_cupy and not is_sparse:
        cupy = pytest.importorskip("cupy")
        X, y = to_dask_cupy_array_xy(X, y, cupy)

    lr = LogisticRegression(fit_intercept=fit_intercept)
    lr.fit(X, y)
    lr.predict(X)
    lr.predict_proba(X)


@pytest.mark.parametrize("fit_intercept", [True, False])
@pytest.mark.parametrize(
    "is_sparse,is_cupy", [(True, False), (False, False), (False, True)]
)
def test_lm(fit_intercept, is_sparse, is_cupy):
    _maybe_skip_sparse_error(fit_intercept, is_sparse, is_cupy)

    X, y = make_regression(
        n_samples=100, n_features=5, chunksize=10, is_sparse=is_sparse
    )
    if is_cupy and not is_sparse:
        cupy = pytest.importorskip("cupy")
        X, y = to_dask_cupy_array_xy(X, y, cupy)
    lr = LinearRegression(fit_intercept=fit_intercept)
    lr.fit(X, y)
    lr.predict(X)
    if fit_intercept:
        assert lr.intercept_ is not None


@pytest.mark.parametrize("fit_intercept", [True, False])
@pytest.mark.parametrize(
    "is_sparse,is_cupy", [(True, False), (False, False), (False, True)]
)
def test_big(fit_intercept, is_sparse, is_cupy):
    _maybe_skip_sparse_error(fit_intercept, is_sparse, is_cupy)

    with dask.config.set(scheduler="synchronous"):
        X, y = make_classification(is_sparse=is_sparse)
        if is_cupy and not is_sparse:
            cupy = pytest.importorskip("cupy")
            X, y = to_dask_cupy_array_xy(X, y, cupy)
        lr = LogisticRegression(fit_intercept=fit_intercept)
        lr.fit(X, y)
        lr.predict(X)
        lr.predict_proba(X)
    if fit_intercept:
        assert lr.intercept_ is not None


@pytest.mark.parametrize("fit_intercept", [True, False])
@pytest.mark.parametrize(
    "is_sparse,is_cupy", [(True, False), (False, False), (False, True)]
)
def test_poisson_fit(fit_intercept, is_sparse, is_cupy):
    _maybe_skip_sparse_error(fit_intercept, is_sparse, is_cupy)

    with dask.config.set(scheduler="synchronous"):
        X, y = make_poisson(is_sparse=is_sparse)
        if is_cupy and not is_sparse:
            cupy = pytest.importorskip("cupy")
            X, y = to_dask_cupy_array_xy(X, y, cupy)
        pr = PoissonRegression(fit_intercept=fit_intercept)
        pr.fit(X, y)
        pr.predict(X)
        pr.get_deviance(X, y)
    if fit_intercept:
        assert pr.intercept_ is not None


def test_in_pipeline():
    from sklearn.pipeline import make_pipeline

    X, y = make_classification(n_samples=100, n_features=5, chunksize=10)
    pipe = make_pipeline(DoNothingTransformer(), LogisticRegression())
    pipe.fit(X, y)


def test_gridsearch():
    from sklearn.pipeline import make_pipeline

    dcv = pytest.importorskip("dask_searchcv")

    X, y = make_classification(n_samples=100, n_features=5, chunksize=10)
    grid = {"logisticregression__lamduh": [0.001, 0.01, 0.1, 0.5]}
    pipe = make_pipeline(DoNothingTransformer(), LogisticRegression())
    search = dcv.GridSearchCV(pipe, grid, cv=3)
    search.fit(X, y)



import dask.array as da
import numpy as np
import sparse

from dask_glm.utils import exp


def make_classification(
    n_samples=1000,
    n_features=100,
    n_informative=2,
    scale=1.0,
    chunksize=100,
    is_sparse=False,
):
    """
    Generate a dummy dataset for classification tasks.

    Parameters
    ----------
    n_samples : int
        number of rows in the output array
    n_features : int
        number of columns (features) in the output array
    n_informative : int
        number of features that are correlated with the outcome
    scale : float
        Scale the true coefficient array by this
    chunksize : int
        Number of rows per dask array block.
    is_sparse: bool
        Return a sparse matrix

    Returns
    -------
    X : dask.array, size ``(n_samples, n_features)``
    y : dask.array, size ``(n_samples,)``
        boolean-valued array

    Examples
    --------
    >>> X, y = make_classification()
    >>> X
    dask.array<da.random.normal, shape=(1000, 100), dtype=float64, chunksize=(100, 100)>
    >>> y
    dask.array<lt, shape=(1000,), dtype=bool, chunksize=(100,)>
    """
    X = da.random.normal(
        0, 1, size=(n_samples, n_features), chunks=(chunksize, n_features)
    )
    if is_sparse:
        X = X.map_blocks(sparse.COO)
    informative_idx = np.random.choice(n_features, n_informative)
    beta = (np.random.random(n_features) - 1) * scale
    z0 = X[:, informative_idx].dot(beta[informative_idx])
    y = da.random.random(z0.shape, chunks=(chunksize,)) < 1 / (1 + da.exp(-z0))
    return X, y


def make_regression(
    n_samples=1000,
    n_features=100,
    n_informative=2,
    scale=1.0,
    chunksize=100,
    is_sparse=False,
):
    """
    Generate a dummy dataset for regression tasks.

    Parameters
    ----------
    n_samples : int
        number of rows in the output array
    n_features : int
        number of columns (features) in the output array
    n_informative : int
        number of features that are correlated with the outcome
    scale : float
        Scale the true coefficient array by this
    chunksize : int
        Number of rows per dask array block.
    is_sparse: bool
        Return a sparse matrix

    Returns
    -------
    X : dask.array, size ``(n_samples, n_features)``
    y : dask.array, size ``(n_samples,)``
        real-valued array

    Examples
    --------
    >>> X, y = make_regression()
    >>> X
    dask.array<da.random.normal, shape=(1000, 100), dtype=float64, chunksize=(100, 100)>
    >>> y
    dask.array<da.random.random_sample, shape=(1000,), dtype=float64, chunksize=(100,)>
    """
    X = da.random.normal(
        0, 1, size=(n_samples, n_features), chunks=(chunksize, n_features)
    )
    if is_sparse:
        X = X.map_blocks(sparse.COO)
    informative_idx = np.random.choice(n_features, n_informative)
    beta = (np.random.random(n_features) - 1) * scale
    z0 = X[:, informative_idx].dot(beta[informative_idx])
    y = da.random.random(z0.shape, chunks=(chunksize,))
    return X, y


def make_poisson(
    n_samples=1000,
    n_features=100,
    n_informative=2,
    scale=1.0,
    chunksize=100,
    is_sparse=False,
):
    """
    Generate a dummy dataset for modeling count data.

    Parameters
    ----------
    n_samples : int
        number of rows in the output array
    n_features : int
        number of columns (features) in the output array
    n_informative : int
        number of features that are correlated with the outcome
    scale : float
        Scale the true coefficient array by this
    chunksize : int
        Number of rows per dask array block.
    is_sparse: bool
        Return a sparse matrix

    Returns
    -------
    X : dask.array, size ``(n_samples, n_features)``
    y : dask.array, size ``(n_samples,)``
        array of non-negative integer-valued data

    Examples
    --------
    >>> X, y = make_classification()
    >>> X
    dask.array<da.random.normal, shape=(1000, 100), dtype=float64, chunksize=(100, 100)>
    >>> y
    dask.array<da.random.poisson, shape=(1000,), dtype=int64, chunksize=(100,)>
    """
    X = da.random.normal(
        0, 1, size=(n_samples, n_features), chunks=(chunksize, n_features)
    )
    if is_sparse:
        X = X.map_blocks(sparse.COO)
    informative_idx = np.random.choice(n_features, n_informative)
    beta = (np.random.random(n_features) - 1) * scale
    z0 = X[:, informative_idx].dot(beta[informative_idx])
    rate = exp(z0)
    y = da.random.poisson(rate, size=1, chunks=(chunksize,))
    return X, y



import dask.array as da
import numpy as np
import pytest
from dask import persist

from dask_glm.algorithms import admm, local_update
from dask_glm.families import Logistic, Normal
from dask_glm.regularizers import L1
from dask_glm.utils import make_y, to_dask_cupy_array_xy


@pytest.mark.parametrize("N", [1000, 10000])
@pytest.mark.parametrize(
    "beta",
    [
        np.array([-1.5, 3]),
        np.array([35, 2, 0, -3.2]),
        np.array([-1e-2, 1e-4, 1.0, 2e-3, -1.2]),
    ],
)
@pytest.mark.parametrize("family", [Logistic, Normal])
def test_local_update(N, beta, family):
    M = beta.shape[0]
    X = np.random.random((N, M))
    y = np.random.random(N) > 0.4
    u = np.zeros(M)
    z = np.random.random(M)
    rho = 1e7

    def create_local_gradient(func):
        def wrapped(beta, X, y, z, u, rho):
            return func(beta, X, y) + rho * (beta - z + u)

        return wrapped

    def create_local_f(func):
        def wrapped(beta, X, y, z, u, rho):
            return func(beta, X, y) + (rho / 2) * np.dot(beta - z + u, beta - z + u)

        return wrapped

    f = create_local_f(family.pointwise_loss)
    fprime = create_local_gradient(family.pointwise_gradient)

    result = local_update(X, y, beta, z, u, rho, f=f, fprime=fprime)

    assert np.allclose(result, z, atol=2e-3)


@pytest.mark.parametrize("N", [1000, 10000])
@pytest.mark.parametrize("nchunks", [5, 10])
@pytest.mark.parametrize("p", [1, 5, 10])
@pytest.mark.parametrize("is_cupy", [True, False])
def test_admm_with_large_lamduh(N, p, nchunks, is_cupy):
    X = da.random.random((N, p), chunks=(N // nchunks, p))
    beta = np.random.random(p)
    y = make_y(X, beta=np.array(beta), chunks=(N // nchunks,))

    if is_cupy:
        cupy = pytest.importorskip("cupy")
        X, y = to_dask_cupy_array_xy(X, y, cupy)

    X, y = persist(X, y)
    z = admm(X, y, regularizer=L1(), lamduh=1e5, rho=20, max_iter=500)

    assert np.allclose(z, np.zeros(p), atol=1e-4)



import dask.array as da
import numpy as np
import pytest
import sparse
from dask.array.utils import assert_eq

from dask_glm import utils


def test_normalize_normalizes():
    @utils.normalize
    def do_nothing(X, y):
        return np.array([0.0, 1.0, 2.0])

    X = da.from_array(np.array([[1, 0, 0], [1, 2, 2]]), chunks=(2, 3))
    y = da.from_array(np.array([0, 1, 0]), chunks=(3,))
    res = do_nothing(X, y)
    np.testing.assert_equal(res, np.array([-3.0, 1.0, 2.0]))


def test_normalize_doesnt_normalize():
    @utils.normalize
    def do_nothing(X, y):
        return np.array([0.0, 1.0, 2.0])

    X = da.from_array(np.array([[1, 0, 0], [1, 2, 2]]), chunks=(2, 3))
    y = da.from_array(np.array([0, 1, 0]), chunks=(3,))
    res = do_nothing(X, y, normalize=False)
    np.testing.assert_equal(res, np.array([0, 1, 2]))


def test_normalize_normalizes_if_intercept_not_present():
    @utils.normalize
    def do_nothing(X, y):
        return np.array([0.0, 1.0, 2.0])

    X = da.from_array(np.array([[1, 0, 0], [3, 9.0, 2]]), chunks=(2, 3))
    y = da.from_array(np.array([0, 1, 0]), chunks=(3,))
    res = do_nothing(X, y)
    np.testing.assert_equal(res, np.array([0, 1 / 4.5, 2]))


def test_normalize_raises_if_multiple_constants():
    @utils.normalize
    def do_nothing(X, y):
        return np.array([0.0, 1.0, 2.0])

    X = da.from_array(np.array([[1, 2, 3], [1, 2, 3]]), chunks=(2, 3))
    y = da.from_array(np.array([0, 1, 0]), chunks=(3,))
    with pytest.raises(ValueError):
        do_nothing(X, y)


def test_add_intercept():
    X = np.zeros((4, 4))
    result = utils.add_intercept(X)
    expected = np.array(
        [
            [0, 0, 0, 0, 1],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 0, 1],
        ],
        dtype=X.dtype,
    )
    assert_eq(result, expected)


def test_add_intercept_dask():
    X = da.from_array(np.zeros((4, 4)), chunks=(2, 4))
    result = utils.add_intercept(X)
    expected = da.from_array(
        np.array(
            [
                [0, 0, 0, 0, 1],
                [0, 0, 0, 0, 1],
                [0, 0, 0, 0, 1],
                [0, 0, 0, 0, 1],
            ],
            dtype=X.dtype,
        ),
        chunks=2,
    )
    assert_eq(result, expected)


def test_add_intercept_sparse():
    X = sparse.COO(np.zeros((4, 4)))
    result = utils.add_intercept(X)
    expected = sparse.COO(
        np.array(
            [
                [0, 0, 0, 0, 1],
                [0, 0, 0, 0, 1],
                [0, 0, 0, 0, 1],
                [0, 0, 0, 0, 1],
            ],
            dtype=X.dtype,
        )
    )
    assert (result == expected).all()


@pytest.mark.xfail(
    reason=(
        "TODO: ValueError: This operation requires consistent fill-values, "
        "but argument 1 had a fill value of 1.0, which is different from a "
        "fill_value of 0.0 in the first argument."
    )
)
def test_add_intercept_sparse_dask():
    X = da.from_array(sparse.COO(np.zeros((4, 4))), chunks=(2, 4))
    result = utils.add_intercept(X)
    expected = da.from_array(
        sparse.COO(
            np.array(
                [
                    [0, 0, 0, 0, 1],
                    [0, 0, 0, 0, 1],
                    [0, 0, 0, 0, 1],
                    [0, 0, 0, 0, 1],
                ],
                dtype=X.dtype,
            )
        ),
        chunks=2,
    )
    assert_eq(result, expected)


def test_sparse():
    x = sparse.COO({(0, 0): 1, (1, 2): 2, (2, 1): 3})
    y = x.todense()
    assert utils.sum(x) == utils.sum(x.todense())
    for func in [utils.sigmoid, utils.sum, utils.exp]:
        assert (func(x) == func(y)).all()


def test_dask_array_is_sparse():
    assert utils.is_dask_array_sparse(da.from_array(sparse.COO([], [], shape=(10, 10))))
    assert utils.is_dask_array_sparse(da.from_array(sparse.eye(10)))
    assert not utils.is_dask_array_sparse(da.from_array(np.eye(10)))


@pytest.mark.xfail(
    reason="dask does not forward DOK in _meta "
    "(https://github.com/pydata/sparse/issues/292)"
)
def test_dok_dask_array_is_sparse():
    assert utils.is_dask_array_sparse(da.from_array(sparse.DOK((10, 10))))


def test_dot_with_cupy():
    cupy = pytest.importorskip("cupy")

    # dot(cupy.array, cupy.array)
    A = cupy.random.rand(100, 100)
    B = cupy.random.rand(100)
    ans = cupy.dot(A, B)
    res = utils.dot(A, B)
    assert_eq(ans, res)

    # dot(dask.array, cupy.array)
    dA = da.from_array(A, chunks=(10, 100))
    res = utils.dot(dA, B).compute()
    assert_eq(ans, res)

    # dot(cupy.array, dask.array)
    dB = da.from_array(B, chunks=(10))
    res = utils.dot(A, dB).compute()
    assert_eq(ans, res)


def test_dot_with_sparse():
    A = sparse.random((1024, 64))
    B = sparse.random((64))
    ans = sparse.dot(A, B)

    # dot(sparse.array, sparse.array)
    res = utils.dot(A, B)
    assert_eq(ans, res)

    # dot(sparse.array, dask.array)
    res = utils.dot(A, da.from_array(B, chunks=B.shape))
    assert_eq(ans, res.compute())

    # dot(dask.array, sparse.array)
    res = utils.dot(da.from_array(A, chunks=A.shape), B)
    assert_eq(ans, res.compute())



from __future__ import absolute_import, division, print_function

from dask_glm.utils import dot, exp, log1p, sigmoid


class Logistic(object):
    """Implements methods for `Logistic regression`_,

    Useful for classifying binary outcomes.

    .. _Logistic regression: https://en.wikipedia.org/wiki/Logistic_regression
    """

    @staticmethod
    def loglike(Xbeta, y):
        """
        Evaluate the logistic loglikeliehood

        Parameters
        ----------
        Xbeta : array, shape (n_samples, n_features)
        y : array, shape (n_samples)
        """
        enXbeta = exp(-Xbeta)
        return (Xbeta + log1p(enXbeta)).sum() - dot(y, Xbeta)

    @staticmethod
    def pointwise_loss(beta, X, y):
        """Logistic Loss, evaluated point-wise."""
        beta, y = beta.ravel(), y.ravel()
        Xbeta = X.dot(beta)
        return Logistic.loglike(Xbeta, y)

    @staticmethod
    def pointwise_gradient(beta, X, y):
        """Logistic gradient, evaluated point-wise."""
        beta, y = beta.ravel(), y.ravel()
        Xbeta = X.dot(beta)
        return Logistic.gradient(Xbeta, X, y)

    @staticmethod
    def gradient(Xbeta, X, y):
        """Logistic gradient"""
        p = sigmoid(Xbeta)
        return dot(X.T, p - y)

    @staticmethod
    def hessian(Xbeta, X):
        """Logistic hessian"""
        p = sigmoid(Xbeta)
        return dot(p * (1 - p) * X.T, X)


class Normal(object):
    """Implements methods for `Linear regression`_,

    Useful for modeling continuous outcomes.

    .. _Linear regression: https://en.wikipedia.org/wiki/Linear_regression
    """

    @staticmethod
    def loglike(Xbeta, y):
        return ((y - Xbeta) ** 2).sum()

    @staticmethod
    def pointwise_loss(beta, X, y):
        beta, y = beta.ravel(), y.ravel()
        Xbeta = X.dot(beta)
        return Normal.loglike(Xbeta, y)

    @staticmethod
    def pointwise_gradient(beta, X, y):
        beta, y = beta.ravel(), y.ravel()
        Xbeta = X.dot(beta)
        return Normal.gradient(Xbeta, X, y)

    @staticmethod
    def gradient(Xbeta, X, y):
        return 2 * dot(X.T, Xbeta) - 2 * dot(X.T, y)

    @staticmethod
    def hessian(Xbeta, X):
        return 2 * dot(X.T, X)


class Poisson(object):
    """
    This implements `Poisson regression`_

    Useful for modelling count data.

    .. _Poisson regression: https://en.wikipedia.org/wiki/Poisson_regression
    """

    @staticmethod
    def loglike(Xbeta, y):
        eXbeta = exp(Xbeta)
        yXbeta = y * Xbeta
        return (eXbeta - yXbeta).sum()

    @staticmethod
    def pointwise_loss(beta, X, y):
        beta, y = beta.ravel(), y.ravel()
        Xbeta = X.dot(beta)
        return Poisson.loglike(Xbeta, y)

    @staticmethod
    def pointwise_gradient(beta, X, y):
        beta, y = beta.ravel(), y.ravel()
        Xbeta = X.dot(beta)
        return Poisson.gradient(Xbeta, X, y)

    @staticmethod
    def gradient(Xbeta, X, y):
        eXbeta = exp(Xbeta)
        return dot(X.T, eXbeta - y)

    @staticmethod
    def hessian(Xbeta, X):
        eXbeta = exp(Xbeta)
        x_diag_eXbeta = eXbeta[:, None] * X
        return dot(X.T, x_diag_eXbeta)
